(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
    typeof define === 'function' && define.amd ? define(factory) :
    (global = global || self, global.HUI = factory());
}(this, function () { 'use strict';

    var _document = document,
        _createTextNode = _document.createTextNode.bind(_document),
        _Object = Object,
        _Array = Array,
        _Symbol = Symbol,
        SYMBOL_ITERATOR = _Symbol.iterator,
        _Boolean = Boolean,
        _Map = Map,
        _Node = Node,
        _console = console,
        _now = Date.now,
        _requestAnimationFrame = requestAnimationFrame,
        _assign = _Object.assign,
        _is = _Object.is,
        _keys = _Object.keys,
        _entries = _Object.entries,
        _create = _Object.create,
        _Object$prototype = _Object.prototype,
        _toString = _Object$prototype.toString,
        _hasOwnProperty = _Object$prototype.hasOwnProperty,
        _isArray = _Array.isArray,
        _from = _Array.from,
        _Array$prototype = _Array.prototype,
        _splice = _Array$prototype.splice,
        _push = _Array$prototype.push,
        _indexOf = _Array$prototype.indexOf,
        _undefined = undefined,
        _null = null,
        _Infinity = Infinity;

    var registry = new _Map();
    var define = function define(name, desc) {
      var type = _Symbol(name);

      registry.set(type, desc);
      return type;
    };

    var toArr = function toArr(a) {
      return _isArray(a) ? a : [a];
    };
    var toFrag = function toFrag(nodes) {
      return nodes.reduce(function (frag, node) {
        return frag.appendChild(node), frag;
      }, _document.createDocumentFragment());
    };
    var isHNode = function isHN(value) {
      return value && isObject(value) && value.isHN;
    };
    var replaceNodes = function replNodes(ownerNode, oldNodes, newNodes) {
      oldNodes.forEach(function (oldNode, i) {
        if (i) {
          ownerNode.removeChild(oldNode);
        } else {
          ownerNode.replaceChild(toFrag(newNodes), oldNode);
        }
      });
    };
    var inherit = function inherit(target, source) {
      target.nodes = source.nodes;
      target.owner = source.owner;
      target.ownNode = source.ownNode;
      target.ctx = source.ctx;
      target.sto = source.sto;

      if (source.desc) {
        target.out = source.out;
      } else {
        target.evMap = source.evMap;
      }
    };
    var supply = function supply(target, defaults) {
      _keys(defaults).forEach(function (key) {
        if (!_hasOwnProperty.call(target, key)) {
          target[key] = defaults[key];
        }
      });
    };
    var isObject = function isObj(value) {
      return typeof value === 'object';
    };

    var renderCallbacks = new Array();
    var renderToDOM = function render(src, options) {
      if (options === void 0) {
        options = _create(_null);
      }

      var _options = options,
          _options$parent = _options.parent,
          parent = _options$parent === void 0 ? _document.body : _options$parent,
          clear = _options.clear,
          owner = _options.owner,
          _options$context = _options.context,
          context = _options$context === void 0 ? createStore() : _options$context,
          defaultContext = _options.defaultContext,
          contextHandlers = _options.contextHandlers;

      if (defaultContext) {
        context.setSome(defaultContext);
      }

      if (contextHandlers) {
        context.handleSome(contextHandlers);
      }

      if (options.sync) {
        if (clear) {
          _from(parent.childNodes).forEach(function (childNode) {
            parent.removeChild(childNode);
          });
        }

        parent.appendChild(toFrag(toNodeArr(src, context, parent, owner)));
      } else {
        renderCallbacks.push(function () {
          renderToDOM(src, {
            parent: parent,
            owner: owner,
            clear: clear,
            context: context,
            sync: true
          });
        });

        if (!willTick) {
          reqTick();
        }
      }
    };

    var noCmpProps = ['children'];
    var patch = function patchEle(element, hNode, curProps, oldProps, curPropKeys) {
      var curProp, oldProp;
      curPropKeys.forEach(function (key) {
        curProp = curProps[key];
        oldProp = oldProps[key];

        if (noCmpProps.includes(key) || !HUI.cmp(curProp, oldProp)) {
          handleProp(element, hNode, key, curProp, oldProp);
        }
      });
    };

    var clear = function clr(hNode) {
      var desc = hNode.desc,
          props = hNode.props;
      hNode.active = false;

      if (desc) {
        var sto = hNode.sto,
            ctx = hNode.ctx;

        try {
          if (hNode.eff) {
            hNode.eff.forEach(function (effCanceller) {
              effCanceller();
            });
          }

          if (desc.clear) {
            desc.clear.call(hNode, props, sto, ctx);
          }
        } catch (err) {
          _console.error(err);
        }
      } else {
        if (props.ref) {
          props.ref();
        }
      }

      var index = expired.indexOf(hNode);

      if (~index) {
        expired.splice(index, 1);
      }

      if (hNode.out) {
        hNode.out.forEach(function (child) {
          if (isHNode(child)) {
            clear(child);
          }
        });
      }
    };

    var updateChildren = function updChd(element, hNode, newChildren, oldChildren) {
      var childNodes = element.childNodes,
          oldChildrenCount = oldChildren.length;
      var nodeOffset = 0,
          oldProps,
          oldNodes,
          newChild,
          newNodes,
          newProps,
          newPropKeys;
      oldChildren.forEach(function (oldChild, i) {
        newChild = newChildren[i];

        if (newChild) {
          if (isHNode(oldChild)) {
            oldNodes = oldChild.nodes;

            if (isHNode(newChild) && newChild.type === oldChild.type) {
              newNodes = oldNodes;

              if (oldChild.desc) {
                inherit(newChild, oldChild);
                nodeOffset += newNodes.length;
                return mark(newChild);
              } else {
                newPropKeys = _keys(newProps = newChild.props);

                if (_keys(oldProps = oldChild.props).every(function (key) {
                  return newPropKeys.includes(key);
                })) {
                  inherit(newChild, oldChild);
                  nodeOffset++;
                  return patch(newNodes[0], newChild, newProps, oldProps, newPropKeys);
                }
              }
            }

            clear(oldChild);
          } else if (HUI.cmp(oldChild, newChild)) {
            return nodeOffset++;
          } else {
            oldNodes = childNodes.length ? [childNodes[nodeOffset]] : [];
          }

          nodeOffset += (newNodes = toNodeArr(newChild, hNode.ctx, element, hNode)).length;

          if (oldNodes.length) {
            replaceNodes(element, oldNodes, newNodes);
          } else {
            element.appendChild(toFrag(newNodes));
          }
        } else {
          if (isHNode(oldChild)) {
            clear(oldChild);
            oldNodes = oldChild.nodes;
            oldChild.nodes.forEach(function (node) {
              element.removeChild(node);
            });
            oldChild.nodes.length = 0;
          } else if (childNodes.length > nodeOffset) {
            element.removeChild(childNodes[nodeOffset++]);
          }
        }
      });

      if (oldChildrenCount < newChildren.length) {
        renderToDOM(newChildren.slice(oldChildrenCount), {
          parent: element,
          owner: hNode,
          context: hNode.ctx
        });
      }
    };

    var propHandlers = new _Map([['children', function chdHandler(element, newChildren, oldChildren, hNode) {
      updateChildren(element, hNode, newChildren, oldChildren || []);
    }], ['style', function styHandler(element, style) {
      if (style && isObject(style)) {
        _assign(element.style, style);
      } else {
        try {
          element.style = style;
        } catch (_unused) {
          element.setAttribute('style', style);
        }
      }
    }], ['class', function clsHandler(element, classes) {
      element.setAttribute('class', _isArray(classes) ? classes.filter(_Boolean).join(' ') : classes);
    }], ['ref', function refHandler(element, callback) {
      callback(element);
    }], ['attr', function attrHandler(element, attributes) {
      _entries(attributes).forEach(function (pair) {
        element.setAttribute(pair[0], pair[1]);
      });
    }], ['prop', function propHandler(element, properties) {
      _assign(element, properties);
    }]]);

    var listen = function listen(element, event, listener) {
      var capture = false,
          passive = true,
          once = false;

      for (var i = 0; i < 3; i++) {
        if (event.endsWith('Capture')) {
          capture = true;
          event = event.slice(0, -7);
        } else if (event.endsWith('Nonpassive')) {
          passive = false;
          event = event.slice(0, -10);
        } else if (event.endsWith('Once')) {
          once = true;
          event = event.slice(0, -4);
        }
      }

      var options = (capture || !passive || once) && {
        capture: capture,
        passive: passive,
        once: once
      };
      element.addEventListener(event, listener, options);
      return [event, listener, options];
    };

    var handleProp = function handleProp(element, hNode, key, newValue, oldValue) {
      var handler = propHandlers.get(key);

      if (handler) {
        handler(element, newValue, oldValue, hNode);
      } else {
        if (key.startsWith('on')) {
          var evMap = hNode.evMap,
              record = evMap.get(key);

          if (record) {
            element.removeEventListener(record[0], record[1], record[2]);
          }

          evMap.set(key, listen(element, key.slice(2), newValue));
        } else if (key in element) {
          try {
            element[key] = newValue;
          } catch (_unused) {
            element.setAttribute(key, newValue);
          }
        } else {
          element.setAttribute(key, newValue);
        }
      }
    };

    var handleError = function handleErr(err, hNode) {
      var owner = hNode.owner,
          desc;

      if (owner) {
        while (!((desc = owner.desc) && desc.catch) && owner.owner) {
          owner = owner.owner;
        }

        if (owner.desc) {
          owner.err = err;
          mark(owner);
          return;
        }
      }

      throw err;
    };

    var initComponent = function initCom(hNode, store) {
      var props = hNode.props,
          ctx = hNode.ctx,
          desc = hNode.desc,
          defaultProps = desc.defaultProps,
          defaultStore = desc.defaultStore,
          storeHandlers = desc.storeHandlers,
          state = desc.state,
          context = desc.context,
          init = desc.init,
          effects = desc.effects;

      if (defaultProps) {
        supply(props, defaultProps);
      }

      hNode.sto = store;

      if (defaultStore) {
        store.setSome(defaultStore);
      }

      if (state) {
        hNode.sto.bind(hNode, state);
      }

      if (storeHandlers) {
        store.handleSome(storeHandlers);
      }

      if (context) {
        ctx.bind(hNode, context);
      }

      if (init) {
        init.call(hNode, props, store, ctx);
      }

      if (effects) {
        hNode.eff = effects.map(function (effect) {
          return effect.call(hNode, props, store, ctx);
        }).filter(_Boolean);
      }
    };

    var toNodeArr = function toNodes(src, context, ownerNode, owner) {
      var srcType = typeof src;

      if (srcType === 'string') {
        return [_createTextNode(src)];
      } else if (srcType === 'number') {
        return [_createTextNode(String(src))];
      }

      if (src) {
        if (_isArray(src) && src.length) {
          return src.flat(_Infinity).flatMap(function (s) {
            return toNodeArr(s, context, ownerNode, owner);
          });
        } else if (isHNode(src)) {
          var type = src.type,
              desc = src.desc;
          src.ownNode = ownerNode;
          src.owner = owner;
          src.ctx = context;

          if (desc) {
            var store = createStore();

            try {
              src.active = false;
              initComponent(src, store);
              return src.nodes = toNodeArr(src.out = toArr(desc.render.call(src, src.props, store, context)).flat(_Infinity), context, ownerNode, src);
            } catch (err) {
              if (desc.catch) {
                return src.nodes = toNodeArr(src.out = toArr(desc.catch.call(src, err, src.props, store, context)).flat(_Infinity), context, ownerNode, src);
              } else {
                handleError(err, src);
              }
            } finally {
              src.active = true;
            }
          } else {
            var props = src.props;
            src.evMap = new _Map();
            var element = props.xmlns ? _document.createElementNS(props.xmlns, type) : _document.createElement(type);

            _entries(props).forEach(function (pair) {
              handleProp(element, src, pair[0], pair[1]);
            });

            src.out = [];
            src.nodes = [element];
            return [element];
          }
        }
      }

      return [_createTextNode('')];
    };

    var updateComponent = function updCom(hNode) {
      var desc = hNode.desc,
          out = hNode.out,
          nodes = hNode.nodes,
          owner = hNode.owner,
          ownNode = hNode.ownNode,
          ctx = hNode.ctx,
          props = hNode.props,
          sto = hNode.sto,
          err = hNode.err,
          outLength = out.length,
          ownerNodes = owner && owner.nodes;
      var newNodes = [].concat(nodes);

      try {
        hNode.active = false;

        if (err) {
          hNode.err = _undefined;
          throw err;
        }

        var ownerNodeNodes = ownNode.childNodes;
        var old,
            oldProps,
            oldNodes,
            oldNodesLength,
            curNodes,
            curProps,
            curPropKeys,
            nodeOffset = 0,
            nextNodeIndex,
            nextNode;
        var newOut = hNode.out = toArr(desc.render.call(hNode, props, sto, ctx)).flat(_Infinity),
            newOutLength = newOut.length;
        newOut.forEach(function (cur, i) {
          if (i < outLength) {
            old = out[i];
            oldNodesLength = 1;

            if (isHNode(old)) {
              oldNodesLength = (oldNodes = old.nodes).length;
              clear(old);

              if (isHNode(cur) && old.type === cur.type) {
                if (old.desc) {
                  inherit(cur, old);
                  mark(cur);
                  return nodeOffset += oldNodesLength;
                } else {
                  curPropKeys = _keys(curProps = cur.props);

                  if (_keys(oldProps = old.props).every(function (k) {
                    return curPropKeys.includes(k);
                  })) {
                    nodeOffset++;
                    inherit(cur, old);
                    return patch(oldNodes[0], cur, curProps, oldProps, curPropKeys);
                  }
                }
              }
            } else if (!isHNode(cur) && HUI.cmp(old, cur)) {
              return nodeOffset++;
            }

            curNodes = toNodeArr(cur, ctx, ownNode, hNode);
            oldNodes = _splice.apply(newNodes, [newNodes.indexOf(nodes[nodeOffset]), oldNodesLength].concat(curNodes));
            nodeOffset += oldNodesLength;
            replaceNodes(ownNode, oldNodes, curNodes);
          } else {
            if (i === outLength) {
              nextNodeIndex = _indexOf.call(ownerNodeNodes, newNodes[newNodes.length - 1]) + 1;
              nextNode = ownerNodeNodes[nextNodeIndex] || _null;
            }

            _push.apply(newNodes, curNodes = toNodeArr(cur, ctx, ownNode, hNode));

            ownNode.insertBefore(toFrag(curNodes), nextNode);
          }
        });

        if (outLength > newOutLength) {
          out.slice(newOutLength).forEach(function (extra) {
            if (isHNode(extra)) {
              clear(extra);
            }
          });
          newNodes.splice(nodeOffset).forEach(function (extraNode) {
            ownNode.removeChild(extraNode);
          });
        }
      } catch (err) {
        if (desc.catch) {
          newNodes = toNodeArr(hNode.out = toArr(desc.catch.call(hNode, err, props, sto, ctx)), ctx, ownNode, hNode);
          replaceNodes(ownNode, nodes, newNodes);
          out.forEach(function (child) {
            if (isHNode(child)) {
              clear(child);
            }
          });
        } else {
          handleError(err, hNode);
        }
      } finally {
        hNode.active = true;
      }

      if (ownerNodes) {
        var firstNode = nodes[0],
            restArgs = [nodes.length].concat(newNodes);
        var target = owner,
            targetNodes = ownerNodes;

        while (target) {
          targetNodes = target.nodes;

          _splice.apply(targetNodes, [targetNodes.indexOf(firstNode)].concat(restArgs));

          target = target.owner;
        }
      }

      hNode.nodes = newNodes;
    };

    var expired = new Array();
    var deferCallbacks = new Array();
    var willTick = false;

    var ticker = function frame() {
      willTick = false;
      var deadline = _now() + HUI.frameLimit;
      var cur, i;

      for (i = 0; i < expired.length; i++) {
        cur = expired[i];

        if (!cur) {
          continue;
        }

        expired[i] = _undefined;

        try {
          updateComponent(cur);
        } catch (err) {
          expired.splice(0, i + 1);
          throw err;
        }

        if (_now() >= deadline) {
          expired.splice(0, i + 1);
          return reqTick();
        }
      }

      expired.length = 0;

      for (i = 0; i < renderCallbacks.length; i++) {
        renderCallbacks[i]();

        if (_now() >= deadline) {
          renderCallbacks.splice(0, i + 1);
          return reqTick();
        }
      }

      renderCallbacks.length = 0;

      for (i = 0; i < deferCallbacks.length; i++) {
        deferCallbacks[i]();

        if (_now() >= deadline) {
          deferCallbacks.splice(0, i + 1);
          return reqTick();
        }
      }

      deferCallbacks.length = 0;
    };

    var reqTick = function reqTick() {
      HUI.tick(ticker);
      willTick = true;
    };
    var mark = function mark(hNode) {
      if (!expired.includes(hNode)) {
        expired.push(hNode);

        if (!willTick) {
          reqTick();
        }
      }
    };
    var defer = function defer(callback) {
      for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      deferCallbacks.push(function () {
        callback.apply(_undefined, args);
      });

      if (!willTick) {
        reqTick();
      }
    };

    var createStore = function crtSto() {
      var valueMap = new _Map(),
          bindingMap = new _Map(),
          setterMap = new _Map(),
          handlerMap = new _Map();
      var store = {
        valueMap: valueMap,
        bindingMap: bindingMap,
        setterMap: setterMap,
        handlerMap: handlerMap,
        bind: function s_bind(hNode, subscriptions) {
          subscriptions.forEach(function (key) {
            if (bindingMap.has(key)) {
              bindingMap.get(key).push(hNode);
            } else {
              bindingMap.set(key, [hNode]);
            }
          });
          return this;
        },
        get: function s_get(key) {
          return valueMap.get(key);
        },
        set: function s_set(key, value, force) {
          if (force || !HUI.cmp(value, store.get(key))) {
            valueMap.set(key, value);

            if (bindingMap.has(key)) {
              bindingMap.get(key).forEach(function (hNode) {
                if (hNode.active) {
                  mark(hNode);
                }
              });
            }
          }

          return store;
        },
        setter: function s_setter(key, force) {
          var index = +!force;

          if (setterMap.has(key)) {
            var setters = setterMap.get(key),
                setter = setters[index];

            if (setter) {
              return setter;
            } else {
              return setters[index] = function setter(value) {
                store.set(key, value, force);
              };
            }
          } else {
            var _setters = new Array();

            setterMap.set(key, _setters);
            return _setters[index] = function setter(value) {
              store.set(key, value, force);
            };
          }
        },
        setSome: function s_setSome(pairs, force) {
          _entries(pairs).forEach(function (pair) {
            store.set(pair[0], pair[1], force);
          });

          return this;
        },
        toggle: function s_toggle(key) {
          return store.set(key, !store.get(key));
        },
        inc: function s_inc(key, addition) {
          if (addition === void 0) {
            addition = 1;
          }

          return store.set(key, store.get(key) + addition);
        },
        push: function s_push(key) {
          for (var _len = arguments.length, items = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
            items[_key - 1] = arguments[_key];
          }

          return store.set(key, store.get(key).concat(items));
        },
        unshift: function s_unshift(key) {
          for (var _len2 = arguments.length, items = new Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
            items[_key2 - 1] = arguments[_key2];
          }

          return store.set(key, items.concat(store.get(key)));
        },
        slice: function s_slice(key, start, end) {
          return store.set(key, store.get(key).slice(start, end));
        },
        splice: function s_splice(key, start, deleteCount) {
          var arr = store.get(key).slice();

          for (var _len3 = arguments.length, items = new Array(_len3 > 3 ? _len3 - 3 : 0), _key3 = 3; _key3 < _len3; _key3++) {
            items[_key3 - 3] = arguments[_key3];
          }

          _splice.apply(arr, [start, deleteCount].concat(items));

          return store.set(key, arr);
        },
        handle: function s_handle(name, handler) {
          if (handler) {
            handlerMap.set(name, handler.bind(store));
          } else {
            handlerMap.delete(name);
          }

          return this;
        },
        handleSome: function s_handleSome(handlers) {
          _entries(handlers).forEach(function (pair) {
            store.handle(pair[0], pair[1]);
          });

          return this;
        },
        getHandler: function s_getHandler(name) {
          return handlerMap.get(name);
        },
        trigger: function s_trigger(name) {
          var handler = handlerMap.get(name);

          if (handler) {
            for (var _len4 = arguments.length, args = new Array(_len4 > 1 ? _len4 - 1 : 0), _key4 = 1; _key4 < _len4; _key4++) {
              args[_key4 - 1] = arguments[_key4];
            }

            return handler.apply(store, args);
          }
        }
      };
      return store;
    };

    var Fragment = define('HUI.Fragment', {
      render: function frag_render(props) {
        return props.children;
      }
    });

    var Portal = define('HUI.Portal', {
      storeHandlers: {
        c: function ptl_c(fragment) {
          var ownNode = fragment.ownNode,
              nodes = fragment.nodes;

          if (nodes) {
            clear(fragment);
            nodes.forEach(function (node) {
              ownNode.removeChild(node);
            });
            nodes.length = 0;
          }
        }
      },
      init: function ptl_init(props, store) {
        var _props$parent = props.parent,
            parent = _props$parent === void 0 ? _document.body : _props$parent,
            fragment = HUI(Fragment, _null, props.children);
        store.set('p', parent);
        store.set('f', fragment);
      },
      render: function ptl_render(props, store, context) {
        var fragment = store.get('f');
        store.trigger('c', fragment);
        renderToDOM(fragment, {
          parent: store.get('p'),
          owner: this,
          context: context
        });
      },
      clear: function ptl_clr(props, store) {
        store.trigger('c', store.get('f'));
      }
    });

    var compare = function cmp(a, b) {
      if (_is(a, b)) {
        return true;
      }

      if (_isArray(a)) {
        return _isArray(b) && a.length === b.length && a.every(function (v, i) {
          return compare(b[i], v);
        });
      }

      if (a && b && isObject(a) && isObject(b) && _toString.call(a) === _toString.call(b)) {
        if (a[SYMBOL_ITERATOR]) {
          return b[SYMBOL_ITERATOR] && compare(_from(a), _from(b));
        } else if (a.isHN) {
          return b.isHN && a.type === b.type && compare(a.props, b.props);
        } else if (!(a instanceof _Node || b instanceof _Node)) {
          var keysA = _keys(a),
              keysB = _keys(b);

          return keysA.length === keysB.length && keysA.every(function (k) {
            return keysB.includes(k) && compare(a[k], b[k]);
          });
        }
      }

      return false;
    };

    var HUI = function HUI(type, props) {
      for (var _len = arguments.length, children = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
        children[_key - 2] = arguments[_key];
      }

      return {
        isHN: true,
        type: type,
        desc: registry.get(type),
        props: _assign({
          children: children.flat(_Infinity)
        }, props),
        active: true
      };
    };
    HUI.registry = registry;
    HUI.define = define;
    HUI.createStore = createStore;
    HUI.propHandlers = propHandlers;
    HUI.noCmpProps = noCmpProps;
    HUI.render = renderToDOM;

    HUI.tick = function tick(callback) {
      _requestAnimationFrame(callback);
    };

    HUI.frameLimit = 15;
    HUI.defer = defer;
    HUI.Portal = Portal;
    HUI.Fragment = Fragment;
    HUI.cmp = compare;

    return HUI;

}));
