import { HNode } from "./HNode";
declare type AssertArray<T> = T extends any[] ? T : never;
declare type MapOf<T> = Map<keyof T, Pick<T, keyof T>>;
export declare type Setter<T = unknown> = (value: T) => void;
export declare type SetterRecord<T = unknown> = [Setter<T> | undefined, Setter<T> | undefined];
export declare type HandlerMap<T extends object = any, H extends object = any> = {
    [K in keyof H]: H[K] extends (this: Store<T, H>, ...args: infer A) => any ? (H[K] extends (this: Store<T, H>, ...args: A) => infer R ? ((this: Store<T, H>, ...args: A) => R) : (never)) : (never);
};
export declare type StoreType<S extends Store> = S extends Store<infer T> ? T : never;
export declare type StoreHandlers<S extends Store> = S extends Store<any, infer H> ? H : never;
export declare type EmptyStore = Store<{}, {}>;
export interface Store<T extends object = any, H extends object = any> {
    valueMap: MapOf<T>;
    bindingMap: Map<keyof T, HNode<any>[]>;
    setterMap: Map<keyof T, SetterRecord<any>>;
    handlerMap: MapOf<HandlerMap<T, H>>;
    bind(hNode: HNode<any>, subscriptions: Array<keyof T>): this;
    get<K extends keyof T>(key: K): T[K] | undefined;
    set<K extends keyof T>(key: K, value: T[K], force?: boolean): this;
    setter<K extends keyof T>(key: K, force?: boolean): Setter<T[K]>;
    setSome(pairs: Partial<T>, forces?: boolean): this;
    toggle(key: keyof T): this;
    inc(key: keyof T, addition?: any): this;
    push<K extends keyof T>(key: K, ...items: AssertArray<T[K]>): this;
    unshift<K extends keyof T>(key: K, ...items: AssertArray<T[K]>): this;
    slice(key: keyof T, start: number, end?: number): this;
    splice<K extends keyof T>(key: K, start: number, deleteCount?: number): this;
    splice<K extends keyof T>(key: K, start: number, deleteCount: number, ...items: AssertArray<T[K]>): this;
    handle<N extends keyof HandlerMap<T, H>>(name: N, handler?: HandlerMap<T, H>[N]): this;
    handleSome(handlers: Partial<HandlerMap<T, H>>): this;
    getHandler<N extends keyof HandlerMap<T, H>>(name: N): HandlerMap<T, H>[N] | undefined;
    trigger<N extends keyof HandlerMap<T, H>>(name: N, ...args: Parameters<HandlerMap<T, H>[N]>): ReturnType<HandlerMap<T, H>[N]> | undefined;
}
export declare const createStore: <T extends object = any, H extends object = any>() => Store<T, H>;
export {};
